package org.ow2.orchestra.designer.wsdlNavigator;

import org.ow2.orchestra.designer.bpmn.model.data.type.DataTypeModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.WSDLMessageModel;
import org.ow2.orchestra.designer.bpmn.model.operations.InterfaceModel;
import org.ow2.orchestra.designer.bpmn.model.operations.OperationModel;
import org.ow2.orchestra.designer.fileNavigator.Navigator;
import org.ow2.orchestra.designer.models.transformers.bpmn2designer.Bpmn2DesignerContext;
import org.ow2.orchestra.designer.models.transformers.bpmn2designer.ElementTransformer;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.wsdl.WsdlInlineUtil;
import org.ow2.orchestra.util.wsdl.WsdlUtil;

import javax.wsdl.Definition;
import javax.wsdl.Import;
import javax.wsdl.Message;
import javax.wsdl.Operation;
import javax.wsdl.PortType;
import javax.wsdl.Types;
import javax.wsdl.extensions.schema.Schema;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

public class WsdlNavigator {

  public static final String WSDL_DIRECTORY_NAME = "OrchestraDesignerWSDLs";

  private final File projectsRootDirectory;

  private static final Logger LOG = Logger.getLogger(WsdlNavigator.class.getName());

  public WsdlNavigator() {
    // get workspace root from navigator
    // so navigator should initialize workspace if needed
    File workspacesRoot = new Navigator().getProjectsRootDirectory();

    // create wsdl repository for the current user
    this.projectsRootDirectory = new File(workspacesRoot, WSDL_DIRECTORY_NAME);
    if (!this.projectsRootDirectory.exists()) {
      if (!this.projectsRootDirectory.mkdirs()) {
        String message = "Unable to create wsdl repository at " + this.projectsRootDirectory.getAbsolutePath();
        LOG.severe(message);
        throw new RuntimeException(message);
      }
    }
  }

  /**
   * @return the list of projects available in the designer.
   */
  public List<Service> getAvailableServices() {
    // List files in the root directory (= projects)
    final File[] wsdlFiles = this.projectsRootDirectory.listFiles();

    final List<Service> services = new ArrayList<Service>();
    Bpmn2DesignerContext context = new Bpmn2DesignerContext();
    for (final File wsdlFile : wsdlFiles) {
      if (!wsdlFile.isHidden()) {
        try {
          Definition wsdlDefinition = WsdlUtil.readWsdl(wsdlFile.toURI().toURL());
          context.getWsdlDefinitions().put(wsdlFile.getName(), wsdlDefinition);
          // Get Schema definitions
          Types types = wsdlDefinition.getTypes();
          if (types != null) {
            int i = 0;
            for (Object extensibilityElement : types.getExtensibilityElements()) {
              i++;
              if (extensibilityElement instanceof Schema) {
                String systemId = wsdlFile.toURI().toURL() + "#types" + i;
                context.getXmlSchemaCollection().read(((Schema) extensibilityElement).getElement(), systemId);
              }
            }
          }
          for (Object portType : wsdlDefinition.getPortTypes().values()) {
            Service service = new Service();
            service.setName(((PortType) portType).getQName().getLocalPart());
            service.setNamespace(((PortType) portType).getQName().getNamespaceURI());
            service.setWsdlPath(wsdlFile.getName());
            InterfaceModel interfaceModel = new InterfaceModel();
            interfaceModel.setId("_" + UUID.randomUUID().toString());
            interfaceModel.setName(((PortType) portType).getQName().getLocalPart());
            interfaceModel.setPortTypeQName(((PortType) portType).getQName().toString());
            service.setInterfaceModel(interfaceModel);
            for (Object operation : ((PortType) portType).getOperations()) {
              Operation op = (Operation) operation;
              OperationModel operationModel = new OperationModel();
              operationModel.setId("_" + UUID.randomUUID().toString());
              operationModel.setName(op.getName());
              Message wsdlInMessage = op.getInput().getMessage();

              operationModel.setInMessageType((WSDLMessageModel) getDataTypeModel(context, wsdlInMessage));
              if (op.getOutput() != null) {
                Message wsdlOutMessage = op.getOutput().getMessage();
                operationModel.setOutMessageType((WSDLMessageModel) getDataTypeModel(context, wsdlOutMessage));
              }
              operationModel.setInterfaceModel(interfaceModel);
              interfaceModel.getOperations().put(op.getName(), operationModel);
            }
            services.add(service);
          }
        } catch (Exception e) {
          Misc.fastDynamicLog(LOG, Level.WARNING, "exception while reading %s: %s", wsdlFile.getPath(), e.getMessage());
        }
      }
    }
    return services;
  }

  private DataTypeModel getDataTypeModel(final Bpmn2DesignerContext context, final Message wsdlMessage) {

    final WSDLMessageModel wsdlMessageModel = new WSDLMessageModel();
    wsdlMessageModel.setQName(wsdlMessage.getQName().toString());

    wsdlMessageModel.setId("_" + UUID.randomUUID().toString());
    ElementTransformer.getWSDLMessagePartModel(context, wsdlMessageModel, wsdlMessage);
    return wsdlMessageModel;
  }


  public boolean uploadWsdl(Map<String, byte[]> files) throws IOException {
    Map<String, Definition> definitions = new HashMap<String, Definition>();
    Map<String, String> definitionAliases = new HashMap<String, String>();
    for (Entry<String, byte[]> file : files.entrySet()) {
      if (file.getKey().endsWith(".wsdl")) {
        // Try to parse wsdl
        Definition wsdlDefinition = WsdlUtil.readWsdl(new Locator(file.getKey(), files));
        // Inline XSD schema definitions
        WsdlInlineUtil.inlineWsdlFile(wsdlDefinition);
        // rename imports and add to definitions
        this.renameWsdlImports(file.getKey(), wsdlDefinition, definitionAliases, definitions);
      }
    }
    for (Entry<String, Definition> file : definitions.entrySet()) {
      Misc.write(WsdlUtil.writeWsdl(file.getValue()), new File(this.projectsRootDirectory, definitionAliases.get(file.getKey())));
    }
    return true;
  }

  public boolean copyWSDL(String wsdlFile, File outputDirectory) throws IOException {
    Map<String, Definition> definitions = new HashMap<String, Definition>();
    Map<String, String> definitionAliases = new HashMap<String, String>();
    // don't rename main wsdl
    definitionAliases.put(wsdlFile, wsdlFile);
    // Try to parse wsdl
    Definition wsdlDefinition = WsdlUtil.readWsdl(new File(projectsRootDirectory, wsdlFile).toURI().toURL());
    // Inline XSD schema definitions
    WsdlInlineUtil.inlineWsdlFile(wsdlDefinition);
    // rename imports and add to definitions
    this.renameWsdlImports(wsdlFile, wsdlDefinition, definitionAliases, definitions);
    for (Entry<String, Definition> file : definitions.entrySet()) {
      Misc.write(WsdlUtil.writeWsdl(file.getValue()), new File(outputDirectory, definitionAliases.get(file.getKey())));
    }
    return true;
  }

  private void renameWsdlImports(String name, Definition wsdlDefinition, Map<String, String> aliases, Map<String, Definition> definitions) {
    if (definitions.containsKey(name)) {
      return;
    }
    // Add to definitions
    definitions.put(name, wsdlDefinition);
    // Generate alias
    if (!aliases.containsKey(name)) {
      aliases.put(name, UUID.randomUUID().toString() + ".wsdl");
    }
    // Process wsdl imports
    final Map<String, List<Import>> imports = wsdlDefinition.getImports();
    if (!imports.isEmpty()) {
      for (final List<Import> l : imports.values()) {
        for (final Import impor : l) {
          // If imported definition is imported from a file,
          // rename the imports of the imported definition
          // and relocate this import with the new alias for the imported definition.
          final Definition importedDefinition = impor.getDefinition();
          String importName = impor.getLocationURI();
          if (!(importName.startsWith("http://") || importName.startsWith("https://"))) {
            if (importName.contains("/")) {
              importName = importName.substring(importName.lastIndexOf("/") + 1);
            }
            this.renameWsdlImports(importName, importedDefinition, aliases, definitions);
            impor.setLocationURI(aliases.get(importName));
          }
        }
      }
    }
  }
}
